/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.pinot.common.function.scalar;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Base64;
import org.apache.pinot.spi.annotations.ScalarFunction;


/**
 * Contains function to convert a datatype to another datatype.
 */
public class DataTypeConversionFunctions {
  private DataTypeConversionFunctions() {

  }

  /**
   * Converts big decimal string representation to bytes.
   * Only scale of upto 2 bytes is supported by the function
   * @param number big decimal number in plain string. e.g. '1234.12121'
   * @return The result byte array contains the bytes of the unscaled value appended to bytes of the scale in BIG ENDIAN order.
   */
  @ScalarFunction
  public static byte[] bigDecimalToBytes(String number) {
    BigDecimal bigDecimal = new BigDecimal(number);
    return bigDecimalToBytes(bigDecimal);
  }

  /**
   * Converts bytes value representation generated by {@link #bigDecimalToBytes(String)} back to string big decimal
   * @param bytes array that contains the bytes of the unscaled value appended to 2 bytes of the scale in BIG ENDIAN order.
   * @return plain string representation of big decimal
   */
  @ScalarFunction
  public static String bytesToBigDecimal(byte[] bytes) {
    BigDecimal number = bytesToBigDecimalObject(bytes);
    return number.toString();
  }

  /**
   * convert simple hex string to byte array
   * @param hex a plain hex string e.g. 'f0e1a3b2'
   * @return byte array representation of hex string
   */
  @ScalarFunction
  public static byte[] hexToBytes(String hex) {
    int len = hex.length();
    byte[] data = new byte[len / 2];
    for (int i = 0; i < len; i += 2) {
      data[i / 2] = (byte) ((Character.digit(hex.charAt(i), 16) << 4) + Character.digit(hex.charAt(i + 1), 16));
    }
    return data;
  }

  /**
   * convert simple bytes array to hex string
   * @param bytes any byte array
   * @return plain hex string e.g. 'f012be3c'
   */
  @ScalarFunction
  public static String bytesToHex(byte[] bytes) {
    StringBuilder sb = new StringBuilder();
    for (byte b : bytes) {
      sb.append(String.format("%02X ", b));
    }

    return sb.toString();
  }

  /**
   * Converts bytes value representation generated by {@link #bigDecimalToBytes(String)} back to string big decimal
   * @param bytes array that contains the bytes of the unscaled value appended to 2 bytes of the scale in BIG ENDIAN order.
   * @return big decimal object
   */
  public static BigDecimal bytesToBigDecimalObject(byte[] bytes) {
    int scale = 0;
    scale += (((int) bytes[0]) << (8));
    scale += (((int) bytes[1]));
    byte[] vals = new byte[bytes.length - 2];
    System.arraycopy(bytes, 2, vals, 0, vals.length);
    BigInteger unscaled = new BigInteger(vals);
    BigDecimal number = new BigDecimal(unscaled, scale);
    return number;
  }

  /**
   * Converts big decimal string representation to bytes.
   * Only scale of upto 2 bytes is supported by the function
   * @param bigDecimal big decimal number object
   * @return The result byte array contains the bytes of the unscaled value appended to bytes of the scale in BIG ENDIAN order.
   */
  public static byte[] bigDecimalToBytes(BigDecimal bigDecimal) {
    int scale = bigDecimal.scale();
    BigInteger unscaled = bigDecimal.unscaledValue();
    byte[] value = unscaled.toByteArray();
    byte[] bigDecimalBytesArray = new byte[value.length + 2];

    bigDecimalBytesArray[0] = (byte) (scale >>> 8);
    bigDecimalBytesArray[1] = (byte) (scale);

    System.arraycopy(value, 0, bigDecimalBytesArray, 2, value.length);
    return bigDecimalBytesArray;
  }

  /**
   * encode byte array to base64 using {@link Base64}
   * @param input original byte array
   * @return base64 encoded byte array
   */
  @ScalarFunction
  public static byte[] base64Encode(byte[] input) {
    return Base64.getEncoder().encodeToString(input).getBytes();
  }

  /**
   * decode base64 encoded string to bytes using {@link Base64}
   * @param input base64 encoded string
   * @return decoded byte array
   */
  @ScalarFunction
  public static byte[] base64Decode(String input) {
    return Base64.getDecoder().decode(input.getBytes());
  }
}
